/*
/
/ sple_topolgy.c - SpatiaLite - spatial extension for SQLite - topological functions
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "sqlite3ext.h"
#include "spatialite.h"

SPATIALITE_DECLARE int
sple_is_closed (sple_geometry_ptr geom)
{
/* 
/ checks if this LINESTRING geometry is a closed one  [first vertex == last vertex]
/ or if all LINESTRINGs within a MULTILINIESTRING are closed
/
/ returns 1 if TRUE
/ 0 if FALSE
/ or -1 if geometry is not a LINESTRING / MULTILINESTRING
*/
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_closed = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_linestring_ptr line = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		if ((line->geometry + 0)->x ==
		    (line->geometry + line->points - 1)->x
		    && (line->geometry + 0)->y ==
		    (line->geometry + line->points - 1)->y)
		    n_closed++;
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points != 0 || n_polygons != 0)
	return -1;
    if (n_linestrings < 1)
	return -1;
    if (n_linestrings == n_closed)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_is_simple (sple_geometry_ptr geom)
{
/* 
/ checks if this GEOMETRY is a simple one  [has no overrlapping points, or self-intersections etc]
/ a collection is simple if every elementar geometry is simple
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int ie;
    int ie2;
    sple_generic_geometry_ptr pGeo;
    sple_generic_geometry_ptr pGeo2;
    sple_linestring_ptr line;
    sple_linestring_ptr line2;
    sple_linestring_ptr ln;
    sple_polygon_ptr polyg;
    sple_point_ptr point;
    sple_point_ptr point2;
    sple_point_ptr pt;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  if (pGeo->type == SPLE_LINESTRING)
	    {
		/* checks if some LINESTRING is not-simple */
		line = (sple_linestring_ptr) (pGeo->geometry);
		if (!sple_is_simple_line (line))
		    return 0;
	    }
      }
    if (geom->entities > 1)
      {
	  /* we need to check if there are anomalies between entities */
	  for (ie = 0; ie < geom->entities; ie++)
	    {
		pGeo = *(geom->geometries + ie);
		line = NULL;
		point = NULL;
		if (pGeo->type == SPLE_POINT)
		    point = (sple_point_ptr) (pGeo->geometry);
		if (pGeo->type == SPLE_LINESTRING)
		    line = (sple_linestring_ptr) (pGeo->geometry);
		if (pGeo->type == SPLE_POLYGON)
		  {
		      polyg = (sple_polygon_ptr) (pGeo->geometry);
		      line = polyg->external;	/* for polygons we check only the exterior ring */
		  }
		for (ie2 = (ie + 1); ie2 < geom->entities; ie2++)
		  {
		      pGeo2 = *(geom->geometries + ie2);
		      line2 = NULL;
		      point2 = NULL;
		      if (pGeo2->type == SPLE_POINT)
			  point2 = (sple_point_ptr) (pGeo2->geometry);
		      if (pGeo2->type == SPLE_LINESTRING)
			  line2 = (sple_linestring_ptr) (pGeo2->geometry);
		      if (pGeo2->type == SPLE_POLYGON)
			{
			    polyg = (sple_polygon_ptr) (pGeo2->geometry);
			    line2 = polyg->external;	/* for polygons we check only the exterior ring */
			}
		      if (point && point2)
			{
			    /* if two points are the same [x,y] geometry cannot be simple */
			    if (point->x == point2->x && point->y == point2->y)
				return 0;
			}
		      if (line && line2)
			{
			    /* we check for consistency of two lines, or two exterior rings, or a line and an exterior ring */
			    if (!sple_are_simple_lines (line, line2))
				return 0;
			}
		      if ((line || line2) && (point || point2))
			{
			    /* we check that this point is not the same of any linestring/border point */
			    if (line)
				ln = line;
			    else
				ln = line2;
			    if (point)
				pt = point;
			    else
				pt = point2;
			    if (sple_point_in_line (pt, ln))
				return 0;
			}
		  }
	    }
      }
    return 1;
}

SPATIALITE_DECLARE int
sple_is_simple_line (sple_linestring_ptr line)
{
/* 
/ checks if this linestring is simple
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int iv;
    int iv2;
    double x1;
    double y1;
    double ox1;
    double oy1;
    double x2;
    double y2;
    double ox2;
    double oy2;
    double point_x;
    double point_y;
    ox1 = DBL_MIN;
    oy1 = DBL_MIN;
    for (iv = 0; iv < line->points; iv++)
      {
	  x1 = (line->geometry + iv)->x;
	  y1 = (line->geometry + iv)->y;
	  ox2 = DBL_MIN;
	  oy2 = DBL_MIN;
	  for (iv2 = iv + 1; iv2 < line->points; iv2++)
	    {
		x2 = (line->geometry + iv2)->x;
		y2 = (line->geometry + iv2)->y;
		if (x1 == x2 && y1 == y2)	/* ouch !!!! - there is a repeated point */
		  {
		      if (iv == 0 && iv2 == (line->points - 1))
			  ;	/* not testing first vs last point; may be legally closed */
		      else
			  return 0;
		  }
		if (ox1 != DBL_MIN && oy1 != DBL_MIN && ox2 != DBL_MIN
		    && oy2 != DBL_MIN)
		  {
		      /* checks for intersections */
		      if (sple_is_intersection
			  (ox1, oy1, x1, y1, ox2, oy2, x2, y2, &point_x,
			   &point_y))
			  return 0;
		  }
		ox2 = x2;
		oy2 = y2;
	    }
	  ox1 = x1;
	  oy1 = y1;
      }
    return 1;
}

SPATIALITE_DECLARE int
sple_are_simple_lines (sple_linestring_ptr line, sple_linestring_ptr line2)
{
/* 
/ checks if this couple of linestring are simple
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int iv;
    int iv2;
    double x1;
    double y1;
    double ox1;
    double oy1;
    double x2;
    double y2;
    double ox2;
    double oy2;
    double point_x;
    double point_y;
    double mbr1_minx;
    double mbr1_miny;
    double mbr1_maxx;
    double mbr1_maxy;
    double mbr2_minx;
    double mbr2_miny;
    double mbr2_maxx;
    double mbr2_maxy;
    if (line->minx == DBL_MAX && line->miny == DBL_MAX && line->maxx == DBL_MIN
	&& line->maxy == DBL_MIN)
	sple_mbr_linestring (line);	/* computing MBR */
    if (line2->minx == DBL_MAX && line2->miny == DBL_MAX
	&& line2->maxx == DBL_MIN && line2->maxy == DBL_MIN)
	sple_mbr_linestring (line2);	/* computing MBR */
/* setting MBR for linestring # 1 */
    mbr1_minx = line->minx;
    mbr1_maxx = line->maxx;
    mbr1_miny = line->miny;
    mbr1_maxy = line->maxy;
/* setting MBR for linestring # 2 */
    mbr2_minx = line2->minx;
    mbr2_maxx = line2->maxx;
    mbr2_miny = line2->miny;
    mbr2_maxy = line2->maxy;
    if (sple_disjoint_mbrs (mbr1_minx, mbr1_miny, mbr1_maxx, mbr1_maxy,
			    mbr2_minx, mbr2_miny, mbr2_maxx, mbr2_maxy))
	return 0;		/* no intersection or overlapping is possible */
    ox1 = DBL_MIN;
    oy1 = DBL_MIN;
    for (iv = 0; iv < line->points; iv++)
      {
	  x1 = (line->geometry + iv)->x;
	  y1 = (line->geometry + iv)->y;
	  ox2 = DBL_MIN;
	  oy2 = DBL_MIN;
	  for (iv2 = 0; iv2 < line2->points; iv2++)
	    {
		x2 = (line2->geometry + iv2)->x;
		y2 = (line2->geometry + iv2)->y;
		if (x1 == x2 && y1 == y2)	/* ouch !!!! - there is an overlapped point */
		    return 0;
		if (ox1 != DBL_MIN && oy1 != DBL_MIN && ox2 != DBL_MIN
		    && oy2 != DBL_MIN)
		  {
		      /* checks for intersections */
		      if (sple_is_intersection
			  (ox1, oy1, x1, y1, ox2, oy2, x2, y2, &point_x,
			   &point_y))
			  return 0;
		  }
		ox2 = x2;
		oy2 = y2;
	    }
	  ox1 = x1;
	  oy1 = y1;
      }
    return 1;
}

SPATIALITE_DECLARE int
sple_point_in_line (sple_point_ptr point, sple_linestring_ptr line)
{
/* 
/ checks if this point is the same of any linestring point
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int iv;
    double mbr1_minx;
    double mbr1_miny;
    double mbr1_maxx;
    double mbr1_maxy;
    double mbr2_minx;
    double mbr2_miny;
    double mbr2_maxx;
    double mbr2_maxy;
    if (line->minx == DBL_MAX && line->miny == DBL_MAX && line->maxx == DBL_MIN
	&& line->maxy == DBL_MIN)
	sple_mbr_linestring (line);	/* computing MBR */
/* setting MBR for point */
    mbr1_minx = point->x;
    mbr1_miny = point->y;
    mbr1_maxx = point->x;
    mbr1_maxy = point->y;
/* setting MBR for linestring */
    mbr2_minx = line->minx;
    mbr2_maxx = line->maxx;
    mbr2_miny = line->miny;
    mbr2_maxy = line->maxy;
    if (sple_disjoint_mbrs (mbr1_minx, mbr1_miny, mbr1_maxx, mbr1_maxy,
			    mbr2_minx, mbr2_miny, mbr2_maxx, mbr2_maxy))
	return 0;		/* no overlapping is possible */
    for (iv = 0; iv < line->points; iv++)
      {
	  if (point->x == (line->geometry + iv)->x &&
	      point->y == (line->geometry + iv)->y)
	      return (1);
      }
    return 0;
}

SPATIALITE_DECLARE int
sple_is_intersection (double ox1, double oy1, double x1, double y1, double ox2,
		      double oy2, double x2, double y2, double *point_x,
		      double *point_y)
{
/* 
/ checks if there is an intersection between two segments
/ [ox1,oy1][x1,y1] is one segmnent
/ [ox2,oy2][x2,y2] is the other one
/ [point_x,point_y] is the intersection point [if any]
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    double mbr1_minx;
    double mbr1_miny;
    double mbr1_maxx;
    double mbr1_maxy;
    double mbr2_minx;
    double mbr2_miny;
    double mbr2_maxx;
    double mbr2_maxy;
    double mbr3_minx;
    double mbr3_miny;
    double mbr3_maxx;
    double mbr3_maxy;
    double a1;
    double b1;
    double c1;
    double a2;
    double b2;
    double c2;
    double m1;
    double m2;
    double p;
    double det_inv;
    *point_x = DBL_MIN;
    *point_y = DBL_MIN;
    if ((ox1 == ox2 && oy1 == oy1) || (x1 == x2 && y1 == y2) ||
	(ox1 == x2 && oy1 == y1) || (x1 == ox2 && y1 == oy2))
	return 0;		/* the segments are consecutives - this is not an intersection */
/* setting MBR for segment # 1 */
    mbr1_minx = ox1;
    if (x1 < mbr1_minx)
	mbr1_minx = x1;
    mbr1_maxx = ox1;
    if (x1 > mbr1_maxx)
	mbr1_maxx = x1;
    mbr1_miny = oy1;
    if (y1 < mbr1_miny)
	mbr1_minx = y1;
    mbr1_maxy = oy1;
    if (y1 > mbr1_maxy)
	mbr1_maxy = y1;
/* setting MBR for segment # 2 */
    mbr2_minx = ox2;
    if (x2 < mbr2_minx)
	mbr2_minx = x2;
    mbr2_maxx = ox2;
    if (x2 > mbr2_maxx)
	mbr2_maxx = x2;
    mbr2_miny = oy2;
    if (y2 < mbr2_miny)
	mbr2_minx = y2;
    mbr2_maxy = oy2;
    if (y2 > mbr2_maxy)
	mbr2_maxy = y2;
    if (sple_disjoint_mbrs (mbr1_minx, mbr1_miny, mbr1_maxx, mbr1_maxy,
			    mbr2_minx, mbr2_miny, mbr2_maxx, mbr2_maxy))
	return 0;		/* an intersection is impossible */
/* computing angular coefficient for line # 1 */
    if ((x1 - ox1) != 0.0)
	m1 = (y1 - oy1) / (x1 - ox1);
    else
	m1 = DBL_MAX;
/* computing angular coefficient for line # 2 */
    if ((x2 - ox2) != 0)
	m2 = (y2 - oy2) / (x2 - ox2);
    else
	m2 = DBL_MAX;
    if (m1 == m2)		/* segments are paralles - no intersection is possible */
	return 0;
/* coumputing parameter for line # 1 */
    if (m1 == DBL_MAX)
	c1 = oy1;
    else
	c1 = (oy1 - m1 * ox1);
/* coumputing parameter for line # 1 */
    if (m2 == DBL_MAX)
	c2 = oy2;
    else
	c2 = (oy2 - m2 * ox2);
    if (m1 == DBL_MAX)
      {
	  *point_x = ox1;
	  p = m2 * ox1;
	  *point_y = p + c2;	/* first segment is exactly vertical */
      }
    else if (m2 == DBL_MAX)
      {
	  *point_x = ox2;
	  p = m1 * ox2;
	  *point_y = p + c1;	/* second segment is exactly vertical */
      }
    else
      {
	  /* computing intersection point between the lines */
	  a1 = m1;
	  a2 = m2;
	  b1 = -1;
	  b2 = -1;
	  det_inv = 1 / (a1 * b2 - a2 * b1);
	  *point_x = (b1 * c2 - b2 * c1) * det_inv;
	  *point_y = (a2 * c1 - a1 * c2) * det_inv;
      }
/* now we need to check if the intersection is within the MBRs of both segments */
    mbr3_minx = *point_x;
    mbr3_miny = *point_y;
    mbr3_maxx = *point_x;
    mbr3_maxy = *point_y;
    if (sple_disjoint_mbrs (mbr1_minx, mbr1_miny, mbr1_maxx, mbr1_maxy,
			    mbr3_minx, mbr3_miny, mbr3_maxx, mbr3_maxy))
	return 0;		/* an intersection is impossible; intersection falls outside MBR of segment # 1 */
    if (sple_disjoint_mbrs (mbr2_minx, mbr2_miny, mbr2_maxx, mbr2_maxy,
			    mbr3_minx, mbr3_miny, mbr3_maxx, mbr3_maxy))
	return 0;		/* an intersection is impossible; intersection falls outside MBR of segment # 1 */
/* all right, the intersection is valid */
    return 1;
}

SPATIALITE_DECLARE int
sple_equal_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		 double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		 double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if those two MBRs are identicals
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    if (mbr1_minx != mbr2_minx)
	return 0;
    if (mbr1_miny != mbr2_miny)
	return 0;
    if (mbr1_maxx != mbr2_maxx)
	return 0;
    if (mbr1_maxy != mbr2_maxy)
	return 0;
    return 1;
}

SPATIALITE_DECLARE int
sple_disjoint_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		    double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		    double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if thore two MBRs are disjoined
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    if (mbr1_minx >= mbr2_maxx)
	return 1;
    if (mbr1_miny >= mbr2_maxy)
	return 1;
    if (mbr1_maxx <= mbr2_minx)
	return 1;
    if (mbr1_maxy <= mbr2_miny)
	return 1;
    if (mbr2_minx >= mbr1_maxx)
	return 1;
    if (mbr2_miny >= mbr1_maxy)
	return 1;
    if (mbr2_maxx <= mbr1_minx)
	return 1;
    if (mbr2_maxy <= mbr1_miny)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_touches_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		   double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		   double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if those two MBRs touches
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    if (mbr1_minx == mbr2_minx)
	return 1;
    if (mbr1_miny == mbr2_miny)
	return 1;
    if (mbr1_maxx == mbr2_maxx)
	return 1;
    if (mbr1_maxy == mbr2_maxy)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_intersects_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		      double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		      double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if those two MBRs intersects
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int ok_1 = 0;
    int ok_2 = 0;
    if (mbr1_minx > mbr2_minx && mbr1_minx < mbr2_maxx)
	ok_1 = 1;
    if (mbr2_minx > mbr1_minx && mbr2_minx < mbr1_maxx)
	ok_1 = 1;
    if (mbr1_miny > mbr2_miny && mbr1_miny < mbr2_maxy)
	ok_2 = 1;
    if (mbr2_miny > mbr1_miny && mbr2_miny < mbr1_maxy)
	ok_2 = 1;
    if (ok_1 && ok_2)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_overlaps_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		    double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		    double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if those two MBRs overlaps
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    if (sple_disjoint_mbrs (mbr1_minx, mbr1_miny, mbr1_maxx, mbr1_maxy,
			    mbr2_minx, mbr2_miny, mbr2_maxx, mbr2_maxy))
	return 0;
    if (mbr1_minx >= mbr2_minx && mbr1_minx <= mbr2_maxx)
	return 1;
    if (mbr1_maxx >= mbr2_minx && mbr1_maxx <= mbr2_maxx)
	return 1;
    if (mbr1_miny >= mbr2_miny && mbr1_miny <= mbr2_maxy)
	return 1;
    if (mbr1_maxy >= mbr2_miny && mbr1_maxy <= mbr2_maxy)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_contains_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		    double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		    double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if MBR-1 completely contains MBR-2
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int ok_1 = 0;
    int ok_2 = 0;
    int ok_3 = 0;
    int ok_4 = 0;
    if (mbr2_minx >= mbr1_minx && mbr2_minx <= mbr1_maxx)
	ok_1 = 1;
    if (mbr2_maxx >= mbr1_minx && mbr2_maxx <= mbr1_maxx)
	ok_2 = 1;
    if (mbr2_miny >= mbr1_miny && mbr2_miny <= mbr1_maxy)
	ok_3 = 1;
    if (mbr2_maxy >= mbr1_miny && mbr2_maxy <= mbr1_maxy)
	ok_4 = 1;
    if (ok_1 && ok_2 && ok_3 && ok_4)
	return 1;
    return 0;
}

SPATIALITE_DECLARE int
sple_within_mbrs (double mbr1_minx, double mbr1_miny, double mbr1_maxx,
		  double mbr1_maxy, double mbr2_minx, double mbr2_miny,
		  double mbr2_maxx, double mbr2_maxy)
{
/* 
/ checks if MBR-2 completely contains MBR-1
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is one MBR
/ [mbr1_minx,mbr1_miny,mbr1_maxx,mbr1_maxy] is the other one
/
/ returns 1 if TRUE
/ 0 if FALSE
*/
    int ok_1 = 0;
    int ok_2 = 0;
    int ok_3 = 0;
    int ok_4 = 0;
    if (mbr1_minx >= mbr2_minx && mbr1_minx <= mbr2_maxx)
	ok_1 = 1;
    if (mbr1_maxx >= mbr2_minx && mbr1_maxx <= mbr2_maxx)
	ok_2 = 1;
    if (mbr1_miny >= mbr2_miny && mbr1_miny <= mbr2_maxy)
	ok_3 = 1;
    if (mbr1_maxy >= mbr2_miny && mbr1_maxy <= mbr2_maxy)
	ok_4 = 1;
    if (ok_1 && ok_2 && ok_3 && ok_4)
	return 1;
    return 0;
}
