/*
/
/ virtualshape.c - main container for VirtualShape - virtual Table extension for SQLite [shapefile]
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2008  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <errno.h>
#include "sqlite3ext.h"
#include "virtualshape.h"

SQLITE_EXTENSION_INIT1
    static int vshp_create (sqlite3 * db, void *pAux, int argc,
			    const char *const *argv, sqlite3_vtab ** ppVTab,
			    char **pzErr);
static int vshp_connect (sqlite3 * db, void *pAux, int argc,
			 const char *const *argv, sqlite3_vtab ** ppVTab,
			 char **pzErr);
static int vshp_best_index (sqlite3_vtab * pVTab, sqlite3_index_info * pIndex);
static int vshp_disconnect (sqlite3_vtab * pVTab);
static int vshp_destroy (sqlite3_vtab * pVTab);
static int vshp_open (sqlite3_vtab * pVTab, sqlite3_vtab_cursor ** ppCursor);
static int vshp_close (sqlite3_vtab_cursor * pCursor);
static int vshp_filter (sqlite3_vtab_cursor * pCursor, int idxNum,
			const char *idxStr, int argc, sqlite3_value ** argv);
static int vshp_next (sqlite3_vtab_cursor * pCursor);
static int vshp_eof (sqlite3_vtab_cursor * pCursor);
static int vshp_column (sqlite3_vtab_cursor * pCursor,
			sqlite3_context * pContext, int column);
static int vshp_rowid (sqlite3_vtab_cursor * pCursor, sqlite_int64 * pRowid);
static int vshp_update (sqlite3_vtab * pVTab, int argc, sqlite3_value ** argv,
			sqlite_int64 * pRowid);
static int vshp_begin (sqlite3_vtab * pVTab);
static int vshp_sync (sqlite3_vtab * pVTab);
static int vshp_commit (sqlite3_vtab * pVTab);
static int vshp_rollback (sqlite3_vtab * pVTab);

static int vshp_open_shp (const char *path, vshp_vtab_ptr pre_vtab, char *buf);
static unsigned short vshp_import_16 (const unsigned char *p,
				      int little_endian);
static unsigned int vshp_import_32 (const unsigned char *p, int little_endian);
static double vshp_import_64 (const unsigned char *p, int little_endian);
static void vshp_export_32 (unsigned char *p, unsigned int value,
			    int little_endian);
static void vshp_export_64 (unsigned char *p, double value, int little_endian);
static vshp_dbf_field_ptr vshp_alloc_dbf_field (char *name,
						unsigned char type,
						int offset,
						unsigned char length,
						unsigned char decimals);
static void vshp_free_dbf_field (vshp_dbf_field_ptr p);
static vshp_dbf_list_ptr vshp_alloc_dbf_list (char *primarykey,
					      char *geocolumn);
static void vshp_free_dbf_list (vshp_dbf_list_ptr p);
static void vshp_add_dbf_field (vshp_dbf_list_ptr list, char *name,
				unsigned char type, int offset,
				unsigned char length, unsigned char decimals);
static void vshp_read_row (vshp_cursor_ptr cursor);

static int is_duplicate_column (vshp_dbf_list_ptr list, char *name);
static int is_reserved_sql_name (char *name);
static int is_reserved_sqlite_name (char *name);

static void vshp_set_pk_value (vshp_vtab_ptr pVtab, int rowid);
static void vshp_set_geometry_value (vshp_vtab_ptr pVtab, void *blob, int size);
static void vshp_set_null_value (vshp_dbf_field_ptr field);
static void vshp_set_str_value (vshp_dbf_field_ptr field, char *value);
static void vshp_set_int_value (vshp_dbf_field_ptr field, int value);
static void vshp_set_double_value (vshp_dbf_field_ptr field, double value);
static void vshp_free_value (vshp_value_ptr value);

static vshp_geometry_ptr vshp_alloc_geometry ();
static void vshp_free_geometry (vshp_geometry_ptr p);
static vshp_point_ptr vshp_alloc_point (double x, double y);
static void vshp_free_point (vshp_point_ptr p);
static vshp_linestring_ptr vshp_alloc_linestring (int vert);
static void vshp_free_linestring (vshp_linestring_ptr p);
static void vshp_free_polygon (vshp_polygon_ptr p);
static vshp_generic_geometry_ptr vshp_alloc_generic_geometry (short type,
							      void *ptr);
static void vshp_free_generic_geometry (vshp_generic_geometry_ptr p);
static void vshp_set_point (vshp_linestring_ptr, int vert, double x, double y);
static vshp_geometry_ptr new_geom_point (vshp_point_ptr point);
static vshp_geometry_ptr new_geom_line (vshp_linestring_ptr line);
static void add_line_to_geom (vshp_geometry_ptr geom, vshp_linestring_ptr line);
static vshp_geometry_ptr new_geom_polyg (vshp_linestring_ptr ring,
					 vshp_polygon_ptr * pg);
static void add_polyg_to_geom (vshp_geometry_ptr geom, vshp_linestring_ptr ring,
			       vshp_polygon_ptr * pg);
static void add_ring_to_polyg (vshp_polygon_ptr polyg,
			       vshp_linestring_ptr ring);
static int is_clockwise_ring (vshp_linestring_ptr ring);
static void vshp_blob_encode (vshp_geometry_ptr geom,
			      unsigned char **buffer, int *size);
static void vshp_mbr_geometry (vshp_geometry_ptr geom);
static void vshp_mbr_linestring (vshp_linestring_ptr line);
static void vshp_mbr_polygon (vshp_polygon_ptr polyg);

struct sqlite3_module my_module;

VIRTUALSHAPE_DECLARE int
sqlite3_extension_init (sqlite3 * db, char **pzErrMsg,
			const sqlite3_api_routines * pApi)
{
    SQLITE_EXTENSION_INIT2 (pApi) my_module.iVersion = 1;
    my_module.xCreate = &vshp_create;
    my_module.xConnect = &vshp_connect;
    my_module.xBestIndex = &vshp_best_index;
    my_module.xDisconnect = &vshp_disconnect;
    my_module.xDestroy = &vshp_destroy;
    my_module.xOpen = &vshp_open;
    my_module.xClose = &vshp_close;
    my_module.xFilter = &vshp_filter;
    my_module.xNext = &vshp_next;
    my_module.xEof = &vshp_eof;
    my_module.xColumn = &vshp_column;
    my_module.xRowid = &vshp_rowid;
    my_module.xUpdate = &vshp_update;
    my_module.xBegin = &vshp_begin;
    my_module.xSync = &vshp_sync;
    my_module.xCommit = &vshp_commit;
    my_module.xRollback = &vshp_rollback;
    my_module.xFindFunction = NULL;
    sqlite3_create_module (db, "VirtualShape", &my_module, NULL);
    return 0;
}

static int
vshp_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
	     sqlite3_vtab ** ppVTab, char **pzErr)
{
/* creates the virtual table connected to some shapefile */
    const char *path = NULL;
    int len;
    char buf[4096];
    char field[128];
    vshp_vtab pre_vtab;
    vshp_dbf_field_ptr pFld;
/* setting initial values for internal struct */
    pre_vtab.path = NULL;
    pre_vtab.fl_shp = NULL;
    pre_vtab.fl_dbf = NULL;
    pre_vtab.dbf = NULL;
/* checking for shapefile PATH */
    if (argc >= 4)
	path = argv[3];
    else
      {
	  *pzErr =
	      sqlite3_mprintf
	      ("[VirtualShape module] CREATE VIRTUAL: no shapefile path was specified");
	  return SQLITE_ERROR;
      }
/* trying to open files etc in order to ensure we actually have a genuine shapefile */
    if (vshp_open_shp (path, &pre_vtab, buf) == 0)
	;
    else
      {
	  *pzErr =
	      sqlite3_mprintf ("[VirtualShape module] CREATE VIRTUAL: %s", buf);
	  return SQLITE_ERROR;
      }
    vshp_vtab_ptr p_vt = (vshp_vtab_ptr) sqlite3_malloc (sizeof (vshp_vtab));
    if (p_vt == NULL)
      {
	  *pzErr =
	      sqlite3_mprintf
	      ("[VirtualShape module] CREATE VIRTUAL: memory allocation error");
	  return SQLITE_ERROR;
      }
    p_vt->pModule = &my_module;
    p_vt->nRef = 0;
    p_vt->zErrMsg = NULL;
    p_vt->db = db;
    len = strlen (path);
    p_vt->path = malloc (len + 1);
    strcpy (p_vt->path, path);
    p_vt->fl_shp = pre_vtab.fl_shp;
    p_vt->fl_shx = pre_vtab.fl_shx;
    p_vt->fl_dbf = pre_vtab.fl_dbf;
    p_vt->shape = pre_vtab.shape;
    p_vt->dbf = pre_vtab.dbf;
    p_vt->buf_dbf = pre_vtab.buf_dbf;
    p_vt->dbf_hdsz = pre_vtab.dbf_hdsz;
    p_vt->dbf_reclen = pre_vtab.dbf_reclen;
    p_vt->buf_shp = pre_vtab.buf_shp;
    p_vt->shp_bfsz = pre_vtab.shp_bfsz;
/* preparing the COLUMNs for this VIRTUAL TABLE */
    strcpy (buf, "CREATE TABLE ");
    strcat (buf, argv[1]);
    strcat (buf, " (");
    strcat (buf, p_vt->dbf->primary_key_name);
    strcat (buf, " INT, ");
    strcat (buf, p_vt->dbf->geometry_name);
    strcat (buf, " BLOB");
    pFld = p_vt->dbf->first;
    while (pFld)
      {
	  vshp_set_null_value (pFld);
	  if (pFld->type == 'N')
	    {
		if (pFld->decimals > 0 || pFld->length > 8)
		    sprintf (field, "%s DOUBLE", pFld->name);
		else
		    sprintf (field, "%s INT", pFld->name);
	    }
	  else
	      sprintf (field, "%s VARCHAR(%d)", pFld->name, pFld->length);
	  strcat (buf, ", ");
	  strcat (buf, field);
	  pFld = pFld->next;
      }
    strcat (buf, ")");
    if (sqlite3_declare_vtab (db, buf) != SQLITE_OK)
      {
	  *pzErr =
	      sqlite3_mprintf
	      ("[VirtualShape module] CREATE VIRTUAL: invalid SQL statement \"%s\"",
	       buf);
	  return SQLITE_ERROR;
      }
    *ppVTab = (sqlite3_vtab *) p_vt;
    return SQLITE_OK;
}

static int
vshp_connect (sqlite3 * db, void *pAux, int argc, const char *const *argv,
	      sqlite3_vtab ** ppVTab, char **pzErr)
{
/* connects the virtual table to some shapefile - simply aliases vshp_create() */
    return vshp_create (db, pAux, argc, argv, ppVTab, pzErr);
}

static int
vshp_best_index (sqlite3_vtab * pVTab, sqlite3_index_info * pIndex)
{
/* best index selection */
    return SQLITE_OK;
}

static int
vshp_disconnect (sqlite3_vtab * pVTab)
{
/* disconnects the virtual table */
    vshp_vtab *p_vt = (vshp_vtab *) pVTab;
    if (p_vt->fl_shp != NULL)
	fclose (p_vt->fl_shp);
    if (p_vt->fl_shx != NULL)
	fclose (p_vt->fl_shx);
    if (p_vt->fl_dbf != NULL)
	fclose (p_vt->fl_dbf);
    if (p_vt->path)
	free (p_vt->path);
    if (p_vt->dbf)
	vshp_free_dbf_list (p_vt->dbf);
    if (p_vt->buf_dbf)
	free (p_vt->buf_dbf);
    if (p_vt->buf_shp)
	free (p_vt->buf_shp);
    sqlite3_free (p_vt);
    return SQLITE_OK;
}

static int
vshp_destroy (sqlite3_vtab * pVTab)
{
/* destroys the virtual table - simply aliases vshp_disconnect() */
    return vshp_disconnect (pVTab);
}

static int
vshp_open (sqlite3_vtab * pVTab, sqlite3_vtab_cursor ** ppCursor)
{
/* opening a new cursor */
    vshp_cursor_ptr cursor =
	(vshp_cursor_ptr) sqlite3_malloc (sizeof (vshp_cursor));
    if (cursor == NULL)
	return SQLITE_ERROR;
    cursor->pVtab = (vshp_vtab_ptr) pVTab;
    cursor->current_row = -1;
    cursor->eof = 0;
    *ppCursor = (sqlite3_vtab_cursor *) cursor;
    vshp_read_row (cursor);
    return SQLITE_OK;
}

static int
vshp_close (sqlite3_vtab_cursor * pCursor)
{
/* closing the cursor */
    sqlite3_free (pCursor);
    return SQLITE_OK;
}

static int
vshp_filter (sqlite3_vtab_cursor * pCursor, int idxNum, const char *idxStr,
	     int argc, sqlite3_value ** argv)
{
/* setting up a cursor filter */
    return SQLITE_OK;
}

static int
vshp_next (sqlite3_vtab_cursor * pCursor)
{
/* fetching a next row from cursor */
    vshp_cursor_ptr cursor = (vshp_cursor_ptr) pCursor;
    vshp_read_row (cursor);
    return SQLITE_OK;
}

static int
vshp_eof (sqlite3_vtab_cursor * pCursor)
{
/* cursor EOF */
    vshp_cursor_ptr cursor = (vshp_cursor_ptr) pCursor;
    return cursor->eof;
}

static int
vshp_column (sqlite3_vtab_cursor * pCursor, sqlite3_context * pContext,
	     int column)
{
/* fetching value for the Nth column */
    int nCol = 2;
    vshp_cursor_ptr cursor = (vshp_cursor_ptr) pCursor;
    if (column == 0)
      {
	  /* the PRIMARY KEY column */
	  sqlite3_result_int (pContext,
			      cursor->pVtab->dbf->pk_value->int_value);
	  return SQLITE_OK;
      }
    if (column == 1)
      {
	  /* the GEOMETRY column */
	  sqlite3_result_blob (pContext,
			       cursor->pVtab->dbf->geometry->blob_value,
			       cursor->pVtab->dbf->geometry->blob_size,
			       SQLITE_STATIC);
	  return SQLITE_OK;
      }
    vshp_dbf_field_ptr pFld = cursor->pVtab->dbf->first;
    while (pFld)
      {
	  if (nCol == column)
	    {
		if (pFld->value == NULL)
		    sqlite3_result_null (pContext);
		else
		  {
		      switch (pFld->value->type)
			{
			case VSHP_INT_VALUE:
			    sqlite3_result_int (pContext,
						pFld->value->int_value);
			    break;
			case VSHP_DOUBLE_VALUE:
			    sqlite3_result_double (pContext,
						   pFld->value->dbl_value);
			    break;
			case VSHP_TEXT_VALUE:
			    sqlite3_result_text (pContext,
						 pFld->value->txt_value,
						 strlen (pFld->value->
							 txt_value),
						 SQLITE_STATIC);
			    break;
			case VSHP_BLOB_VALUE:
			    sqlite3_result_blob (pContext,
						 pFld->value->blob_value,
						 pFld->value->blob_size,
						 SQLITE_STATIC);
			    break;
			default:
			    sqlite3_result_null (pContext);
			    break;
			};
		  }
		break;
	    }
	  nCol++;
	  pFld = pFld->next;
      }
    return SQLITE_OK;
}

static int
vshp_rowid (sqlite3_vtab_cursor * pCursor, sqlite_int64 * pRowid)
{
/* fetching the ROWID */
    vshp_cursor_ptr cursor = (vshp_cursor_ptr) pCursor;
    *pRowid = cursor->pVtab->dbf->pk_value->int_value;
    return SQLITE_OK;
}

static int
vshp_update (sqlite3_vtab * pVTab, int argc, sqlite3_value ** argv,
	     sqlite_int64 * pRowid)
{
/* generic update [INSERT / UPDATE / DELETE */
    return SQLITE_READONLY;
}

static int
vshp_begin (sqlite3_vtab * pVTab)
{
/* BEGIN TRANSACTION */
    return SQLITE_OK;
}

static int
vshp_sync (sqlite3_vtab * pVTab)
{
/* BEGIN TRANSACTION */
    return SQLITE_OK;
}

static int
vshp_commit (sqlite3_vtab * pVTab)
{
/* BEGIN TRANSACTION */
    return SQLITE_OK;
}

static int
vshp_rollback (sqlite3_vtab * pVTab)
{
/* BEGIN TRANSACTION */
    return SQLITE_OK;
}

static int
vshp_open_shp (const char *path, vshp_vtab_ptr pre_vtab, char *errMsg)
{
/* trying to open the shapefile and initial checkings */
    FILE *fl_shx = NULL;
    FILE *fl_shp = NULL;
    FILE *fl_dbf = NULL;
    char xpath[1024];
    int rd;
    unsigned char buf_shx[256];
    long size_shp;
    long size_shx;
    unsigned char *buf_shp = NULL;
    int buf_size = 1024;
    long shape;
    char *sys_err;
    unsigned char bf[1024];
    int dbf_size;
    int dbf_reclen = 0;
    int dbf_recno;
    int off_dbf;
    int ind;
    char field_name[16];
    vshp_dbf_list_ptr dbf_list = NULL;
    *errMsg = '\0';
    sprintf (xpath, "%s.shx", path);
    fl_shx = fopen (xpath, "rb");
    if (!fl_shx)
	goto no_file;
    sprintf (xpath, "%s.shp", path);
    fl_shp = fopen (xpath, "rb");
    if (!fl_shp)
	goto no_file;
    sprintf (xpath, "%s.dbf", path);
    fl_dbf = fopen (xpath, "rb");
    if (!fl_dbf)
	goto no_file;
/* reading SHX file header */
    rd = fread (buf_shx, sizeof (unsigned char), 100, fl_shx);
    if (rd != 100)
	goto error;
    if (vshp_import_32 (buf_shx + 0, VSHP_BIG_ENDIAN) != 9994)	/* checks the SHX magic number */
	goto error;
    size_shx = vshp_import_32 (buf_shx + 24, VSHP_BIG_ENDIAN);
/* reading SHP file header */
    buf_shp = malloc (sizeof (unsigned char) * buf_size);
    rd = fread (buf_shp, sizeof (unsigned char), 100, fl_shp);
    if (rd != 100)
	goto error;
    if (vshp_import_32 (buf_shp + 0, VSHP_BIG_ENDIAN) != 9994)	/* checks the SHP magic number */
	goto error;
    size_shp = vshp_import_32 (buf_shp + 24, VSHP_BIG_ENDIAN);
    shape = vshp_import_32 (buf_shp + 32, VSHP_LITTLE_ENDIAN);
    if (shape == 1 || shape == 11 || shape == 21 ||
	shape == 3 || shape == 13 || shape == 23 ||
	shape == 5 || shape == 15 || shape == 25)
	;
    else
	goto unsupported;
/* reading DBF file header */
    rd = fread (bf, sizeof (unsigned char), 32, fl_dbf);
    if (rd != 32)
	goto error;
    if (*bf != 0x03)		/* checks the DBF magic number */
	goto error;
    dbf_recno = vshp_import_32 (bf + 4, VSHP_LITTLE_ENDIAN);
    dbf_size = vshp_import_16 (bf + 8, VSHP_LITTLE_ENDIAN);
    dbf_reclen = vshp_import_16 (bf + 10, VSHP_LITTLE_ENDIAN);
    dbf_size--;
    off_dbf = 0;
    dbf_list = vshp_alloc_dbf_list ("PKUID", "GEOMETRY");
    for (ind = 32; ind < dbf_size; ind += 32)
      {
	  /* fetches DBF fields definitions */
	  rd = fread (bf, sizeof (unsigned char), 32, fl_dbf);
	  if (rd != 32)
	      goto error;
	  memcpy (field_name, bf, 11);
	  field_name[11] = '\0';
	  vshp_add_dbf_field (dbf_list, field_name, *(bf + 11), off_dbf,
			      *(bf + 16), *(bf + 17));
	  off_dbf += *(bf + 16);
      }
    pre_vtab->shape = shape;
    pre_vtab->fl_shp = fl_shp;
    pre_vtab->fl_shx = fl_shx;
    pre_vtab->fl_dbf = fl_dbf;
    pre_vtab->dbf = dbf_list;
/* saving the SHP buffer */
    pre_vtab->buf_shp = buf_shp;
    pre_vtab->shp_bfsz = buf_size;
/* allocating DBF buffer */
    pre_vtab->buf_dbf = malloc (sizeof (unsigned char) * dbf_reclen);
    pre_vtab->dbf_hdsz = dbf_size + 1;
    pre_vtab->dbf_reclen = dbf_reclen;
    return 0;
  no_file:
/* one of shapefile's files can't be accessed */
    sys_err = strerror (errno);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (errMsg, "unable to open '%s' for reading: %s", xpath, sys_err);
    return -1;
  error:
/* the shapefile is invalid or corrupted */
    vshp_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    fclose (fl_shx);
    fclose (fl_shp);
    fclose (fl_dbf);
    sprintf (errMsg, "invalid or corrupted shapefile '%s'", path);
    return -1;
  unsupported:
/* the shapefile has an unrecognized shape type */
    vshp_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    fclose (fl_shx);
    fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (errMsg, "shapefile '%s' has an unsupported shape type", path);
    return -1;
}

static void
vshp_read_row (vshp_cursor_ptr cursor)
{
/* trying to read a "row" from shapefile */
    unsigned char buf[512];
    int rd;
    long skpos;
    long offset;
    long off_shp;
    int i;
    unsigned char *blob;
    int blob_size;
    vshp_dbf_field_ptr pFld;
    long sz;
    long shape;
    double x;
    double y;
    int points;
    long n;
    long n1;
    int base;
    int start;
    int end;
    int iv;
    int ind;
    vshp_geometry_ptr geom = NULL;
    vshp_point_ptr point = NULL;
    vshp_linestring_ptr line = NULL;
    vshp_polygon_ptr polyg = NULL;
/* positioning and reading the SHX file */
    offset = 100 + ((cursor->current_row + 1) * 8);	/* 100 bytes for the header + current row displacement; each SHX row = 8 bytes */
    skpos = fseek (cursor->pVtab->fl_shx, offset, SEEK_SET);
    if (skpos != 0)
	goto error;
    rd = fread (buf, sizeof (unsigned char), 8, cursor->pVtab->fl_shx);
    if (rd != 8)
	goto error;
    off_shp = vshp_import_32 (buf, VSHP_BIG_ENDIAN);
/* positioning and reading the DBF file */
    offset =
	cursor->pVtab->dbf_hdsz +
	((cursor->current_row + 1) * cursor->pVtab->dbf_reclen);
    skpos = fseek (cursor->pVtab->fl_dbf, offset, SEEK_SET);
    if (skpos != 0)
	goto error;
    rd = fread (cursor->pVtab->buf_dbf, sizeof (unsigned char),
		cursor->pVtab->dbf_reclen, cursor->pVtab->fl_dbf);
    if (rd != cursor->pVtab->dbf_reclen)
	goto error;
/* positioning and reading corresponding SHP entity - geometry */
    offset = off_shp * 2;
    skpos = fseek (cursor->pVtab->fl_shp, offset, SEEK_SET);
    if (skpos != 0)
	goto error;
    rd = fread (buf, sizeof (unsigned char), 12, cursor->pVtab->fl_shp);
    if (rd != 12)
	goto error;
    sz = vshp_import_32 (buf + 4, VSHP_BIG_ENDIAN);
    shape = vshp_import_32 (buf + 8, VSHP_LITTLE_ENDIAN);
    if (shape != cursor->pVtab->shape)
	goto error;
    if ((sz * 2) > cursor->pVtab->shp_bfsz)
      {
	  /* current buffer is too small; we need to allocate a bigger buffer */
	  free (cursor->pVtab->buf_shp);
	  cursor->pVtab->shp_bfsz = sz * 2;
	  cursor->pVtab->buf_shp =
	      malloc (sizeof (unsigned char) * cursor->pVtab->shp_bfsz);
      }
    if (shape == 1 || shape == 11 || shape == 21)
      {
	  /* shape point */
	  rd = fread (cursor->pVtab->buf_shp, sizeof (unsigned char), 16,
		      cursor->pVtab->fl_shp);
	  if (rd != 16)
	      goto error;
	  x = vshp_import_64 (cursor->pVtab->buf_shp, VSHP_LITTLE_ENDIAN);
	  y = vshp_import_64 (cursor->pVtab->buf_shp + 8, VSHP_LITTLE_ENDIAN);
	  point = vshp_alloc_point (x, y);
	  geom = new_geom_point (point);
      }
    if (shape == 3 || shape == 13 || shape == 23)
      {
	  /* shape polyline */
	  rd = fread (cursor->pVtab->buf_shp, sizeof (unsigned char), 32,
		      cursor->pVtab->fl_shp);
	  if (rd != 32)
	      goto error;
	  rd = fread (cursor->pVtab->buf_shp, sizeof (unsigned char),
		      (sz * 2) - 36, cursor->pVtab->fl_shp);
	  if (rd != (sz * 2) - 36)
	      goto error;
	  n = vshp_import_32 (cursor->pVtab->buf_shp, VSHP_LITTLE_ENDIAN);
	  n1 = vshp_import_32 (cursor->pVtab->buf_shp + 4, VSHP_LITTLE_ENDIAN);
	  base = 8 + (n * 4);
	  start = 0;
	  for (ind = 0; ind < n; ind++)
	    {
		if (ind < (n - 1))
		    end =
			vshp_import_32 (cursor->pVtab->buf_shp + 8 +
					((ind + 1) * 4), VSHP_LITTLE_ENDIAN);
		else
		    end = n1;
		points = end - start;
		line = vshp_alloc_linestring (points);
		points = 0;
		for (iv = start; iv < end; iv++)
		  {
		      x = vshp_import_64 (cursor->pVtab->buf_shp + base +
					  (iv * 16), VSHP_LITTLE_ENDIAN);
		      y = vshp_import_64 (cursor->pVtab->buf_shp + base +
					  (iv * 16) + 8, VSHP_LITTLE_ENDIAN);
		      vshp_set_point (line, points, x, y);
		      start++;
		      points++;
		  }
		if (!geom)
		    geom = new_geom_line (line);
		else
		    add_line_to_geom (geom, line);
	    }
      }
    if (shape == 5 || shape == 15 || shape == 25)
      {
	  /* shape polygon */
	  rd = fread (cursor->pVtab->buf_shp, sizeof (unsigned char), 32,
		      cursor->pVtab->fl_shp);
	  if (rd != 32)
	      goto error;
	  rd = fread (cursor->pVtab->buf_shp, sizeof (unsigned char),
		      (sz * 2) - 36, cursor->pVtab->fl_shp);
	  if (rd != (sz * 2) - 36)
	      goto error;
	  n = vshp_import_32 (cursor->pVtab->buf_shp, VSHP_LITTLE_ENDIAN);
	  n1 = vshp_import_32 (cursor->pVtab->buf_shp + 4, VSHP_LITTLE_ENDIAN);
	  base = 8 + (n * 4);
	  start = 0;
	  for (ind = 0; ind < n; ind++)
	    {
		if (ind < (n - 1))
		    end =
			vshp_import_32 (cursor->pVtab->buf_shp + 8 +
					((ind + 1) * 4), VSHP_LITTLE_ENDIAN);
		else
		    end = n1;
		points = end - start;
		line = vshp_alloc_linestring (points);
		points = 0;
		for (iv = start; iv < end; iv++)
		  {
		      x = vshp_import_64 (cursor->pVtab->buf_shp + base +
					  (iv * 16), VSHP_LITTLE_ENDIAN);
		      y = vshp_import_64 (cursor->pVtab->buf_shp + base +
					  (iv * 16) + 8, VSHP_LITTLE_ENDIAN);
		      vshp_set_point (line, points, x, y);
		      start++;
		      points++;
		  }
		if (!geom)
		    geom = new_geom_polyg (line, &polyg);	/* new geometry - new need to allocate a new POLYGON */
		else
		  {
		      if (is_clockwise_ring (line))
			{
			    /* this one is a POLYGON exterior ring - we need to allocate e new POLYGON */
			    add_polyg_to_geom (geom, line, &polyg);
			}
		      else
			{
			    /* adding an interior ring to current POLYGON */
			    add_ring_to_polyg (polyg, line);
			}
		  }
	    }
      }
/* setting up the current row [primary key] */
    (cursor->current_row)++;
    vshp_set_pk_value (cursor->pVtab, cursor->current_row);
    vshp_blob_encode (geom, &blob, &blob_size);
    vshp_free_geometry (geom);
    geom = NULL;
    vshp_set_geometry_value (cursor->pVtab, blob, blob_size);
/* fetching the DBF values */
    pFld = cursor->pVtab->dbf->first;
    while (pFld)
      {
	  memcpy (buf, cursor->pVtab->buf_dbf + pFld->offset + 1, pFld->length);
	  buf[pFld->length] = '\0';
	  if (*buf == '\0')
	      vshp_set_null_value (pFld);
	  else
	    {
		if (pFld->type == 'N')
		  {
		      if (pFld->decimals > 0 || pFld->length > 8)
			  vshp_set_double_value (pFld, atof ((char *) buf));
		      else
			  vshp_set_int_value (pFld, atoi ((char *) buf));
		  }
		else
		  {
		      for (i = strlen ((char *) buf) - 1; i > 1; i--)
			{
			    /* cleaning up trailing spaces */
			    if (buf[i] == ' ')
				buf[i] = '\0';
			    else
				break;
			}
		      vshp_set_str_value (pFld, (char *) buf);
		  }
	    }
	  pFld = pFld->next;
      }
    return;
  error:
    cursor->eof = 1;
}

static unsigned short
vshp_import_16 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[2];
	unsigned short short_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 1);
	  convert.byte[1] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 1);
	  convert.byte[1] = *(p + 0);
      }
#endif
    return convert.short_value;
}

static unsigned int
vshp_import_32 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[4];
	unsigned int int_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 3);
	  convert.byte[1] = *(p + 2);
	  convert.byte[2] = *(p + 1);
	  convert.byte[3] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 3);
	  convert.byte[1] = *(p + 2);
	  convert.byte[2] = *(p + 1);
	  convert.byte[3] = *(p + 0);
      }
#endif
    return convert.int_value;
}

static double
vshp_import_64 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[8];
	double double_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 7);
	  convert.byte[1] = *(p + 6);
	  convert.byte[2] = *(p + 5);
	  convert.byte[3] = *(p + 4);
	  convert.byte[4] = *(p + 3);
	  convert.byte[5] = *(p + 2);
	  convert.byte[6] = *(p + 1);
	  convert.byte[7] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
	  convert.byte[4] = *(p + 4);
	  convert.byte[5] = *(p + 5);
	  convert.byte[6] = *(p + 6);
	  convert.byte[7] = *(p + 7);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
	  convert.byte[4] = *(p + 4);
	  convert.byte[5] = *(p + 5);
	  convert.byte[6] = *(p + 6);
	  convert.byte[7] = *(p + 7);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 7);
	  convert.byte[1] = *(p + 6);
	  convert.byte[2] = *(p + 5);
	  convert.byte[3] = *(p + 4);
	  convert.byte[4] = *(p + 3);
	  convert.byte[5] = *(p + 2);
	  convert.byte[6] = *(p + 1);
	  convert.byte[7] = *(p + 0);
      }
#endif
    return convert.double_value;
}

static void
vshp_export_32 (unsigned char *p, unsigned int value, int little_endian)
{
    union cvt
    {
	unsigned char byte[4];
	unsigned long int_value;
    } convert;
    convert.int_value = value;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[3];
	  *(p + 1) = convert.byte[2];
	  *(p + 2) = convert.byte[1];
	  *(p + 3) = convert.byte[0];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[3];
	  *(p + 1) = convert.byte[2];
	  *(p + 2) = convert.byte[1];
	  *(p + 3) = convert.byte[0];
      }
#endif
}

static void
vshp_export_64 (unsigned char *p, double value, int little_endian)
{
    union cvt
    {
	unsigned char byte[8];
	double double_value;
    } convert;
    convert.double_value = value;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[7];
	  *(p + 1) = convert.byte[6];
	  *(p + 2) = convert.byte[5];
	  *(p + 3) = convert.byte[4];
	  *(p + 4) = convert.byte[3];
	  *(p + 5) = convert.byte[2];
	  *(p + 6) = convert.byte[1];
	  *(p + 7) = convert.byte[0];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
	  *(p + 4) = convert.byte[4];
	  *(p + 5) = convert.byte[5];
	  *(p + 6) = convert.byte[6];
	  *(p + 7) = convert.byte[7];
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
	  *(p + 4) = convert.byte[4];
	  *(p + 5) = convert.byte[5];
	  *(p + 6) = convert.byte[6];
	  *(p + 7) = convert.byte[7];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[7];
	  *(p + 1) = convert.byte[6];
	  *(p + 2) = convert.byte[5];
	  *(p + 3) = convert.byte[4];
	  *(p + 4) = convert.byte[3];
	  *(p + 5) = convert.byte[2];
	  *(p + 6) = convert.byte[1];
	  *(p + 7) = convert.byte[0];
      }
#endif
}

static void
vshp_set_pk_value (vshp_vtab_ptr pVtab, int value)
{
/* assignes an INTEGER value as the PRIMARY KEY value */
    if (pVtab->dbf->pk_value)
	vshp_free_value (pVtab->dbf->pk_value);
    pVtab->dbf->pk_value = malloc (sizeof (vshp_value));
    pVtab->dbf->pk_value->type = VSHP_INT_VALUE;
    pVtab->dbf->pk_value->txt_value = NULL;
    pVtab->dbf->pk_value->int_value = value;
    pVtab->dbf->pk_value->blob_value = NULL;
}

static void
vshp_set_geometry_value (vshp_vtab_ptr pVtab, void *blob, int size)
{
/* assignes an INTEGER value as the GEOMETRY value */
    if (pVtab->dbf->geometry)
	vshp_free_value (pVtab->dbf->geometry);
    pVtab->dbf->geometry = malloc (sizeof (vshp_value));
    pVtab->dbf->geometry->type = VSHP_BLOB_VALUE;
    pVtab->dbf->geometry->txt_value = NULL;
    pVtab->dbf->geometry->blob_value = blob;
    pVtab->dbf->geometry->blob_size = size;
}

static void
vshp_set_null_value (vshp_dbf_field_ptr field)
{
/* assignes a NULL value to some DBF field */
    if (field->value)
	vshp_free_value (field->value);
    field->value = malloc (sizeof (vshp_value));
    field->value->type = VSHP_NULL_VALUE;
    field->value->txt_value = NULL;
    field->value->blob_value = NULL;
}

static void
vshp_set_int_value (vshp_dbf_field_ptr field, int value)
{
/* assignes an INTEGER value to some DBF field */
    if (field->value)
	vshp_free_value (field->value);
    field->value = malloc (sizeof (vshp_value));
    field->value->type = VSHP_INT_VALUE;
    field->value->txt_value = NULL;
    field->value->blob_value = NULL;
    field->value->int_value = value;
}

static void
vshp_set_double_value (vshp_dbf_field_ptr field, double value)
{
/* assignes a DOUBLE value to some DBF field */
    if (field->value)
	vshp_free_value (field->value);
    field->value = malloc (sizeof (vshp_value));
    field->value->type = VSHP_DOUBLE_VALUE;
    field->value->txt_value = NULL;
    field->value->blob_value = NULL;
    field->value->dbl_value = value;
}

static void
vshp_set_str_value (vshp_dbf_field_ptr field, char *str)
{
/* assignes a STRING value to some DBF field */
    int len = strlen (str);
    if (field->value)
	vshp_free_value (field->value);
    field->value = malloc (sizeof (vshp_value));
    field->value->type = VSHP_TEXT_VALUE;
    field->value->txt_value = malloc (len + 1);
    strcpy (field->value->txt_value, str);
    field->value->blob_value = NULL;
}

static void
vshp_free_value (vshp_value_ptr p)
{
/* frees all memory allocations for this DBF Field value */
    if (!p)
	return;
    if (p->txt_value)
	free (p->txt_value);
    if (p->blob_value)
	free (p->blob_value);
    free (p);
}

static vshp_dbf_field_ptr
vshp_alloc_dbf_field (char *name, unsigned char type,
		      int offset, unsigned char length, unsigned char decimals)
{
/* allocates and initializes a DBF Field definition */
    vshp_dbf_field_ptr p = malloc (sizeof (vshp_dbf_field));
    int len = strlen (name);
    p->name = malloc (len + 1);
    strcpy (p->name, name);
    p->type = type;
    p->offset = offset;
    p->length = length;
    p->decimals = decimals;
    p->value = NULL;
    p->next = NULL;
    return p;
}

static void
vshp_free_dbf_field (vshp_dbf_field_ptr p)
{
/* frees all memory allocations for this DBF Field definition */
    if (!p)
	return;
    if (p->name)
	free (p->name);
    if (p->value != NULL)
	vshp_free_value (p->value);
    free (p);
}

static vshp_dbf_list_ptr
vshp_alloc_dbf_list (char *primarykey, char *geocolumn)
{
/* allocates and initializes the DBF Fields list */
    int len;
    vshp_dbf_list_ptr list = malloc (sizeof (vshp_dbf_list));
    len = strlen (primarykey);
    list->primary_key_name = malloc (len + 1);
    strcpy (list->primary_key_name, primarykey);
    list->pk_value = NULL;
    len = strlen (geocolumn);
    list->geometry_name = malloc (len + 1);
    strcpy (list->geometry_name, geocolumn);
    list->geometry = NULL;
    list->first = NULL;
    list->last = NULL;
    list->seed = 1;
    return list;
}

static void
vshp_free_dbf_list (vshp_dbf_list_ptr list)
{
/* frees all memory allocations related to DBF Fields list */
    vshp_dbf_field_ptr p;
    vshp_dbf_field_ptr pn;
    if (!list)
	return;
    p = list->first;
    while (p)
      {
	  pn = p->next;
	  vshp_free_dbf_field (p);
	  p = pn;
      }
    if (list->primary_key_name)
	free (list->primary_key_name);
    if (list->geometry_name)
	free (list->geometry_name);
    free (p);
}

static void
vshp_add_dbf_field (vshp_dbf_list_ptr list, char *name, unsigned char type,
		    int offset, unsigned char length, unsigned char decimals)
{
/* inserts a Field in the DBF Fields list */
    char alias[64];
    if (!list)
	return;
    strcpy (alias, name);
    while (1)
      {
	  if (is_duplicate_column (list, alias))
	    {
		/* 
		   / the original DBF attribute name causes a non-unique column name; 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  if (is_reserved_sql_name (alias))
	    {
		/* 
		   / the original DBF attribute name is an SQL reserver keyword: 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  if (is_reserved_sqlite_name (alias))
	    {
		/* 
		   / the original DBF attribute name is an SQLite reserver keyword: 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  /* the column name is valid */
	  break;
      }
    vshp_dbf_field_ptr p =
	vshp_alloc_dbf_field (alias, type, offset, length, decimals);
    if (!(list->first))
	list->first = p;
    if (list->last)
	list->last->next = p;
    list->last = p;
}

static int
is_duplicate_column (vshp_dbf_list_ptr list, char *name)
{
/* checks if column-name is non-unique */
    vshp_dbf_field_ptr p = list->first;
    while (p)
      {
	  if (strcasecmp (p->name, name) == 0)
	      return 1;
	  p = p->next;
      }
    if (strcasecmp (list->primary_key_name, name) == 0)
	return 1;
    if (strcasecmp (list->geometry_name, name) == 0)
	return 1;
    return 0;
}

static int
is_reserved_sqlite_name (char *name)
{
/* checks if column-name is an SQLite reserved keyword */
    char *reserved[] = {
	"ALL",
	"ALTER",
	"AND",
	"AS",
	"AUTOINCREMENT",
	"BETWEEN",
	"BY",
	"CASE",
	"CHECK",
	"COLLATE",
	"COMMIT",
	"CONSTRAINT",
	"CREATE",
	"CROSS",
	"DEFAULT",
	"DEFERRABLE",
	"DELETE",
	"DISTINCT",
	"DROP",
	"ELSE",
	"ESCAPE",
	"EXCEPT",
	"FOREIGN",
	"FROM",
	"FULL",
	"GLOB",
	"GROUP",
	"HAVING",
	"IN",
	"INDEX",
	"INNER",
	"INSERT",
	"INTERSECT",
	"INTO",
	"IS",
	"ISNULL",
	"JOIN",
	"LEFT",
	"LIKE",
	"LIMIT",
	"NATURAL",
	"NOT",
	"NOTNULL",
	"NULL",
	"ON",
	"OR",
	"ORDER",
	"OUTER",
	"PRIMARY",
	"REFERENCES",
	"RIGHT",
	"ROLLBACK",
	"SELECT",
	"SET",
	"TABLE",
	"THEN",
	"TO",
	"TRANSACTION",
	"UNION",
	"UNIQUE",
	"UPDATE",
	"USING",
	"VALUES",
	"WHEN",
	"WHERE",
	NULL
    };
    char **pw = reserved;
    while (*pw != NULL)
      {
	  if (strcasecmp (name, *pw) == 0)
	      return 1;
	  pw++;
      }
    return 0;
}

static int
is_reserved_sql_name (char *name)
{
/* checks if column-name is an SQL reserved keyword */
    char *reserved[] = {
	"ABSOLUTE",
	"ACTION",
	"ADD",
	"AFTER",
	"ALL",
	"ALLOCATE",
	"ALTER",
	"AND",
	"ANY",
	"ARE",
	"ARRAY",
	"AS",
	"ASC",
	"ASENSITIVE",
	"ASSERTION",
	"ASYMMETRIC",
	"AT",
	"ATOMIC",
	"AUTHORIZATION",
	"AVG",
	"BEFORE",
	"BEGIN",
	"BETWEEN",
	"BIGINT",
	"BINARY",
	"BIT",
	"BIT_LENGTH",
	"BLOB",
	"BOOLEAN",
	"BOTH",
	"BREADTH",
	"BY",
	"CALL",
	"CALLED",
	"CASCADE",
	"CASCADED",
	"CASE",
	"CAST",
	"CATALOG",
	"CHAR",
	"CHARACTER",
	"CHARACTER_LENGTH",
	"CHAR_LENGTH",
	"CHECK",
	"CLOB",
	"CLOSE",
	"COALESCE",
	"COLLATE",
	"COLLATION",
	"COLUMN",
	"COMMIT",
	"CONDITION",
	"CONNECT",
	"CONNECTION",
	"CONSTRAINT",
	"CONSTRAINTS",
	"CONSTRUCTOR",
	"CONTAINS",
	"CONTINUE",
	"CONVERT",
	"CORRESPONDING",
	"COUNT",
	"CREATE",
	"CROSS",
	"CUBE",
	"CURRENT",
	"CURRENT_DATE",
	"CURRENT_DEFAULT_TRANSFORM_GROUP",
	"CURRENT_PATH",
	"CURRENT_ROLE",
	"CURRENT_TIME",
	"CURRENT_TIMESTAMP",
	"CURRENT_TRANSFORM_GROUP_FOR_TYPE",
	"CURRENT_USER",
	"CURSOR",
	"CYCLE",
	"DATA",
	"DATE",
	"DAY",
	"DEALLOCATE",
	"DEC",
	"DECIMAL",
	"DECLARE",
	"DEFAULT",
	"DEFERRABLE",
	"DEFERRED",
	"DELETE",
	"DEPTH",
	"DEREF",
	"DESC",
	"DESCRIBE",
	"DESCRIPTOR",
	"DETERMINISTIC",
	"DIAGNOSTICS",
	"DISCONNECT",
	"DISTINCT",
	"DO",
	"DOMAIN",
	"DOUBLE",
	"DROP",
	"DYNAMIC",
	"EACH",
	"ELEMENT",
	"ELSE",
	"ELSEIF",
	"END",
	"EQUALS",
	"ESCAPE",
	"EXCEPT",
	"EXCEPTION",
	"EXEC",
	"EXECUTE",
	"EXISTS",
	"EXIT",
	"external",
	"EXTRACT",
	"FALSE",
	"FETCH",
	"FILTER",
	"FIRST",
	"FLOAT",
	"FOR",
	"FOREIGN",
	"FOUND",
	"FREE",
	"FROM",
	"FULL",
	"FUNCTION",
	"GENERAL",
	"GET",
	"GLOBAL",
	"GO",
	"GOTO",
	"GRANT",
	"GROUP",
	"GROUPING",
	"HANDLER",
	"HAVING",
	"HOLD",
	"HOUR",
	"IDENTITY",
	"IF",
	"IMMEDIATE",
	"IN",
	"INDICATOR",
	"INITIALLY",
	"INNER",
	"INOUT",
	"INPUT",
	"INSENSITIVE",
	"INSERT",
	"INT",
	"INTEGER",
	"INTERSECT",
	"INTERVAL",
	"INTO",
	"IS",
	"ISOLATION",
	"ITERATE",
	"JOIN",
	"KEY",
	"LANGUAGE",
	"LARGE",
	"LAST",
	"LATERAL",
	"LEADING",
	"LEAVE",
	"LEFT",
	"LEVEL",
	"LIKE",
	"LOCAL",
	"LOCALTIME",
	"LOCALTIMESTAMP",
	"LOCATOR",
	"LOOP",
	"LOWER",
	"MAP",
	"MATCH",
	"MAX",
	"MEMBER",
	"MERGE",
	"METHOD",
	"MIN",
	"MINUTE",
	"MODIFIES",
	"MODULE",
	"MONTH",
	"MULTISET",
	"NAMES",
	"NATIONAL",
	"NATURAL",
	"NCHAR",
	"NCLOB",
	"NEW",
	"NEXT",
	"NO",
	"NONE",
	"NOT",
	"NULL",
	"NULLIF",
	"NUMERIC",
	"OBJECT",
	"OCTET_LENGTH",
	"OF",
	"OLD",
	"ON",
	"ONLY",
	"OPEN",
	"OPTION",
	"OR",
	"ORDER",
	"ORDINALITY",
	"OUT",
	"OUTER",
	"OUTPUT",
	"OVER",
	"OVERLAPS",
	"PAD",
	"PARAMETER",
	"PARTIAL",
	"PARTITION",
	"PATH",
	"POSITION",
	"PRECISION",
	"PREPARE",
	"PRESERVE",
	"PRIMARY",
	"PRIOR",
	"PRIVILEGES",
	"PROCEDURE",
	"PUBLIC",
	"RANGE",
	"READ",
	"READS",
	"REAL",
	"RECURSIVE",
	"REF",
	"REFERENCES",
	"REFERENCING",
	"RELATIVE",
	"RELEASE",
	"REPEAT",
	"RESIGNAL",
	"RESTRICT",
	"RESULT",
	"RETURN",
	"RETURNS",
	"REVOKE",
	"RIGHT",
	"ROLE",
	"ROLLBACK",
	"ROLLUP",
	"ROUTINE",
	"ROW",
	"ROWS",
	"SAVEPOINT",
	"SCHEMA",
	"SCOPE",
	"SCROLL",
	"SEARCH",
	"SECOND",
	"SECTION",
	"SELECT",
	"SENSITIVE",
	"SESSION",
	"SESSION_USER",
	"SET",
	"SETS",
	"SIGNAL",
	"SIMILAR",
	"SIZE",
	"SMALLINT",
	"SOME",
	"SPACE",
	"SPECIFIC",
	"SPECIFICTYPE",
	"SQL",
	"SQLCODE",
	"SQLERROR",
	"SQLEXCEPTION",
	"SQLSTATE",
	"SQLWARNING",
	"START",
	"STATE",
	"STATIC",
	"SUBMULTISET",
	"SUBSTRING",
	"SUM",
	"SYMMETRIC",
	"SYSTEM",
	"SYSTEM_USER",
	"TABLE",
	"TABLESAMPLE",
	"TEMPORARY",
	"THEN",
	"TIME",
	"TIMESTAMP",
	"TIMEZONE_HOUR",
	"TIMEZONE_MINUTE",
	"TO",
	"TRAILING",
	"TRANSACTION",
	"TRANSLATE",
	"TRANSLATION",
	"TREAT",
	"TRIGGER",
	"TRIM",
	"TRUE",
	"UNDER",
	"UNDO",
	"UNION",
	"UNIQUE",
	"UNKNOWN",
	"UNNEST",
	"UNTIL",
	"UPDATE",
	"UPPER",
	"USAGE",
	"USER",
	"USING",
	"VALUE",
	"VALUES",
	"VARCHAR",
	"VARYING",
	"VIEW",
	"WHEN",
	"WHENEVER",
	"WHERE",
	"WHILE",
	"WINDOW",
	"WITH",
	"WITHIN",
	"WITHOUT",
	"WORK",
	"WRITE",
	"YEAR",
	"ZONE",
	NULL
    };
    char **pw = reserved;
    while (*pw != NULL)
      {
	  if (strcasecmp (name, *pw) == 0)
	      return 1;
	  pw++;
      }
    return 0;
}

static vshp_geometry_ptr
vshp_alloc_geometry ()
{
/* allocates and initializes a GEOMETRY COLLECTION */
    vshp_geometry_ptr p = malloc (sizeof (vshp_geometry));
    p->blob = NULL;
    p->size = 0;
    p->entities = 0;
    p->geometries = NULL;
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

static void
vshp_free_geometry (vshp_geometry_ptr p)
{
/* cleans all memory allocations related to this GEOMETRY COLLECTION */
    int i;
    if (!p)
	return;
    for (i = 0; i < p->entities; i++)
      {
	  if (*(p->geometries + i))
	      vshp_free_generic_geometry (*(p->geometries + i));
      }
    if (p->geometries)
	free (p->geometries);
    free (p);
}

static vshp_point_ptr
vshp_alloc_point (double x, double y)
{
/* allocates and initializes a POINT */
    vshp_point_ptr p = malloc (sizeof (vshp_point));
    p->x = x;
    p->y = y;
    return p;
}

static void
vshp_free_point (vshp_point_ptr p)
{
/* cleans all memory allocations related to this POINT */
    if (!p)
	return;
    free (p);
}

static vshp_linestring_ptr
vshp_alloc_linestring (int vert)
{
/* allocates a LINESTRING */
    if (vert <= 0)
	return NULL;
    vshp_linestring_ptr p = malloc (sizeof (vshp_linestring));
    p->points = vert;
    if (vert <= 0)
	p->geometry = NULL;
    else
	p->geometry = malloc (sizeof (vshp_point) * vert);
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

static void
vshp_free_linestring (vshp_linestring_ptr p)
{
/* cleans all memory allocations related to this LINESTRING */
    if (!p)
	return;
    if (p->geometry)
	free (p->geometry);
    free (p);
}

static void
vshp_free_polygon (vshp_polygon_ptr p)
{
/* cleans all memory allocations related to this POLYGON */
    int i;
    if (!p)
	return;
    if (p->external)
	vshp_free_linestring (p->external);
    for (i = 0; i < p->borders; i++)
	vshp_free_linestring (*(p->internals + i));
    if (p->internals)
	free (p->internals);
    free (p);
}

static vshp_generic_geometry_ptr
vshp_alloc_generic_geometry (short type, void *ptr)
{

/* allocates and initializes a generic geometry container */
    vshp_generic_geometry_ptr p = malloc (sizeof (vshp_generic_geometry));
    p->type = type;
    p->geometry = ptr;
    return p;
}

static void
vshp_free_generic_geometry (vshp_generic_geometry_ptr p)
{
/* cleans all memory allocations related to this generic geometry */
    switch (p->type)
      {
      case VSHP_POINT:
	  vshp_free_point ((vshp_point_ptr) (p->geometry));
	  break;
      case VSHP_LINESTRING:
	  vshp_free_linestring ((vshp_linestring_ptr) (p->geometry));
	  break;
      case VSHP_POLYGON:
	  vshp_free_polygon ((vshp_polygon_ptr) (p->geometry));
	  break;
      };
}

static vshp_geometry_ptr
new_geom_point (vshp_point_ptr point)
{
/*
/ allocates and initializes a new geometry - POINT */
    vshp_geometry_ptr geom = vshp_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (vshp_generic_geometry_ptr));
    *(geom->geometries) =
	vshp_alloc_generic_geometry (VSHP_POINT, (void *) point);
    return geom;
}

static vshp_geometry_ptr
new_geom_line (vshp_linestring_ptr line)
{
/*
/ allocates and initializes a new geometry - LINESTRING */
    vshp_geometry_ptr geom = vshp_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (vshp_generic_geometry_ptr));
    *(geom->geometries) =
	vshp_alloc_generic_geometry (VSHP_LINESTRING, (void *) line);
    return geom;
}

static void
add_line_to_geom (vshp_geometry_ptr geom, vshp_linestring_ptr line)
{
/* inserts an other line in this geometry - MULTILINESTRING */
    vshp_generic_geometry_ptr *old_geometries = geom->geometries;
    int ie;
    geom->geometries =
	malloc (sizeof (vshp_generic_geometry_ptr) * (geom->entities + 1));
    for (ie = 0; ie < geom->entities; ie++)
	*(geom->geometries + ie) = *(old_geometries + ie);
    *(geom->geometries + geom->entities) =
	vshp_alloc_generic_geometry (VSHP_LINESTRING, (void *) line);
    (geom->entities)++;
    free (old_geometries);
}

static vshp_geometry_ptr
new_geom_polyg (vshp_linestring_ptr ring, vshp_polygon_ptr * pg)
{
/*
/ allocates and initializes a new geometry - POLYGON */
    vshp_polygon_ptr polyg;
    vshp_geometry_ptr geom = vshp_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (vshp_generic_geometry_ptr));
    polyg = malloc (sizeof (vshp_polygon));
    *pg = polyg;
    polyg->external = ring;
    polyg->borders = 0;
    polyg->internals = NULL;
    *(geom->geometries) =
	vshp_alloc_generic_geometry (VSHP_POLYGON, (void *) polyg);
    return geom;
}

static void
add_polyg_to_geom (vshp_geometry_ptr geom, vshp_linestring_ptr ring,
		   vshp_polygon_ptr * pg)
{
/* inserts an other line in this geometry - MULTIPOLYGON */
    vshp_polygon_ptr polyg;
    vshp_generic_geometry_ptr *old_geometries = geom->geometries;
    int ie;
    geom->geometries =
	malloc (sizeof (vshp_generic_geometry_ptr) * (geom->entities + 1));
    for (ie = 0; ie < geom->entities; ie++)
	*(geom->geometries + ie) = *(old_geometries + ie);
    polyg = malloc (sizeof (vshp_polygon));
    *pg = polyg;
    polyg->external = ring;
    polyg->borders = 0;
    polyg->internals = NULL;
    *(geom->geometries + geom->entities) =
	vshp_alloc_generic_geometry (VSHP_POLYGON, (void *) polyg);
    (geom->entities)++;
    free (old_geometries);
}

static void
add_ring_to_polyg (vshp_polygon_ptr polyg, vshp_linestring_ptr ring)
{
/* adds an interior ring to last POLYGON of this geometry */
    int ib;
    vshp_linestring_ptr *old_internals = NULL;
    if (!(polyg->borders))
      {
	  /* this one is the first interior ring */
	  polyg->internals = malloc (sizeof (vshp_linestring_ptr));
	  *(polyg->internals) = ring;
	  polyg->borders = 1;
      }
    else
      {
	  /* adding another interior ring */
	  old_internals = polyg->internals;
	  polyg->internals =
	      malloc (sizeof (vshp_linestring_ptr) * (polyg->borders + 1));
	  for (ib = 0; ib < polyg->borders; ib++)
	      *(polyg->internals + ib) = *(old_internals + ib);
	  *(polyg->internals + polyg->borders) = ring;
	  (polyg->borders)++;
	  free (old_internals);
      }
}

static int
is_clockwise_ring (vshp_linestring_ptr ring)
{
/* checks if this ring is clockwise or anti-clockwise */
    int iv;
    int ix;
    double xx;
    double yy;
    double x;
    double y;
    double area = 0.0;
    for (iv = 0; iv < ring->points; iv++)
      {
	  xx = (ring->geometry + iv)->x;
	  yy = (ring->geometry + iv)->y;
	  ix = (iv + 1) % ring->points;
	  x = (ring->geometry + ix)->x;
	  y = (ring->geometry + ix)->y;
	  area += ((xx * y) - (x * yy));
      }
    area /= 2.0;
    if (area >= 0.0)
	return 0;
    else
	return 1;
}

static void
vshp_set_point (vshp_linestring_ptr p, int vert, double x, double y)
{
/* sets the coordinates for a LINESTRING's vertex */
    if (p == NULL)
	return;
    if (vert >= 0 && vert < p->points)
      {
	  (p->geometry + vert)->x = x;
	  (p->geometry + vert)->y = y;
      }
}

static void
vshp_blob_encode (vshp_geometry_ptr geom, unsigned char **result, int *size)
{
/* 
/ builds the BLOB representation for this GEOMETRY
 */
    int ie;
    int ib;
    int iv;
    int entities = 0;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    int type;
    unsigned char *ptr;
    vshp_generic_geometry_ptr pGeo;
    vshp_point_ptr point = NULL;
    vshp_linestring_ptr line = NULL;
    vshp_polygon_ptr polyg = NULL;
    vshp_mbr_geometry (geom);
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* how many entities, and of what kind, do we have ? */
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case VSHP_POINT:
		point = (vshp_point_ptr) (pGeo->geometry);
		entities++;
		n_points++;
		break;
	    case VSHP_LINESTRING:
		line = (vshp_linestring_ptr) (pGeo->geometry);
		entities++;
		n_linestrings++;
		break;
	    case VSHP_POLYGON:
		polyg = (vshp_polygon_ptr) (pGeo->geometry);
		entities++;
		n_polygons++;
		break;
	    };
      }
/* ok, we can determine the geometry class */
    if (n_points == 1 && n_linestrings == 0 && n_polygons == 0)
	type = VSHP_POINT;
    else if (n_points > 1 && n_linestrings == 0 && n_polygons == 0)
	type = VSHP_MULTIPOINT;
    else if (n_points == 0 && n_linestrings == 1 && n_polygons == 0)
	type = VSHP_LINESTRING;
    else if (n_points == 0 && n_linestrings > 1 && n_polygons == 0)
	type = VSHP_MULTILINESTRING;
    else if (n_points == 0 && n_linestrings == 0 && n_polygons == 1)
	type = VSHP_POLYGON;
    else if (n_points == 0 && n_linestrings == 0 && n_polygons > 1)
	type = VSHP_MULTIPOLYGON;
    else
	type = VSHP_GEOMETRYCOLLECTION;
/* and now we compute the size of BLOB */
    *size = 44;			/* header size */
    if (type == VSHP_POINT)
	*size += (sizeof (double) * 2);	/* two points */
    else if (type == VSHP_LINESTRING)
	*size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] for each vertex */
    else if (type == VSHP_POLYGON)
      {
	  line = polyg->external;
	  *size += (8 + ((sizeof (double) * 2) * line->points));	/* # rings + # points + [x.y] array - exterior ring */
	  for (ib = 0; ib < polyg->borders; ib++)
	    {
		line = *(polyg->internals + ib);
		*size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] array - interior ring */
	    }
      }
    else
      {
	  /* this one is not a simple geometry; should be a MULTIxxxx or a GEOMETRYCOLLECTION */
	  *size += 4;		/* # entities */
	  for (ie = 0; ie < geom->entities; ie++)
	    {
		*size += 5;	/* entity header */
		pGeo = *(geom->geometries + ie);
		switch (pGeo->type)
		  {
		  case VSHP_POINT:
		      *size += (sizeof (double) * 2);	/* two points */
		      break;
		  case VSHP_LINESTRING:
		      line = (vshp_linestring_ptr) (pGeo->geometry);
		      *size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] for each vertex */
		      break;
		  case VSHP_POLYGON:
		      polyg = (vshp_polygon_ptr) (pGeo->geometry);
		      line = polyg->external;
		      *size += (8 + ((sizeof (double) * 2) * line->points));	/* # rings + # points + [x.y] array - exterior ring */
		      for (ib = 0; ib < polyg->borders; ib++)
			{
			    line = *(polyg->internals + ib);
			    *size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] array - interior ring */
			}
		      break;
		  };
	    }
      }
    *result = malloc (*size);
    ptr = *result;
/* and finally we build the BLOB */
    if (type == VSHP_POINT)
      {
	  *ptr = VSHP_MARK_START;	/* START signatue */
	  *(ptr + 1) = VSHP_LITTLE_ENDIAN;	/* byte ordering */
	  vshp_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  vshp_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  vshp_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  vshp_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  vshp_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = VSHP_MARK_MBR;	/* MBR signature */
	  vshp_export_32 (ptr + 39, VSHP_POINT, 1);	/* class POINT */
	  vshp_export_64 (ptr + 43, point->x, 1);	/* X */
	  vshp_export_64 (ptr + 51, point->y, 1);	/* Y */
	  *(ptr + 59) = VSHP_MARK_END;	/* END signature */
      }
    else if (type == VSHP_LINESTRING)
      {
	  *ptr = VSHP_MARK_START;	/* START signatue */
	  *(ptr + 1) = VSHP_LITTLE_ENDIAN;	/* byte ordering */
	  vshp_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  vshp_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  vshp_export_64 (ptr + 16, geom->miny, 1);	/* MBR - minimun Y */
	  vshp_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  vshp_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = VSHP_MARK_MBR;	/* MBR signature */
	  vshp_export_32 (ptr + 39, VSHP_LINESTRING, 1);	/* class LINESTRING */
	  vshp_export_32 (ptr + 43, line->points, 1);	/* # points */
	  ptr += 47;
	  for (iv = 0; iv < line->points; iv++)
	    {
		vshp_export_64 (ptr, (line->geometry + iv)->x, 1);
		vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);
		ptr += 16;
	    }
	  *ptr = VSHP_MARK_END;	/* END signature */
      }
    else if (type == VSHP_POLYGON)
      {
	  *ptr = VSHP_MARK_START;	/* START signatue */
	  vshp_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  *(ptr + 1) = VSHP_LITTLE_ENDIAN;	/* byte ordering */
	  vshp_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  vshp_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  vshp_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  vshp_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = VSHP_MARK_MBR;	/* MBR signature */
	  vshp_export_32 (ptr + 39, VSHP_POLYGON, 1);	/* class POLYGON */
	  vshp_export_32 (ptr + 43, polyg->borders + 1, 1);	/* # rings */
	  line = polyg->external;
	  vshp_export_32 (ptr + 47, line->points, 1);	/* # points - exterior ring */
	  ptr += 51;
	  for (iv = 0; iv < line->points; iv++)
	    {
		vshp_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - exterior ring */
		vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - exterior ring */
		ptr += 16;
	    }
	  for (ib = 0; ib < polyg->borders; ib++)
	    {
		line = *(polyg->internals + ib);
		vshp_export_32 (ptr, line->points, 1);	/* # points - interior ring */
		ptr += 4;
		for (iv = 0; iv < line->points; iv++)
		  {
		      vshp_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - interior ring */
		      vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - interior ring */
		      ptr += 16;
		  }
	    }
	  *ptr = VSHP_MARK_END;	/* END signature */
      }
    else
      {
	  /* this one is a MULTIxxxx or a GEOMETRYCOLLECTION - building the main header */
	  *ptr = VSHP_MARK_START;	/* START signatue */
	  *(ptr + 1) = VSHP_LITTLE_ENDIAN;	/* byte ordering */
	  vshp_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  vshp_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  vshp_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  vshp_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  vshp_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = VSHP_MARK_MBR;	/* MBR signature */
	  vshp_export_32 (ptr + 39, type, 1);	/* geometric class */
	  vshp_export_32 (ptr + 43, entities, 1);	/* # entities */
	  ptr += 47;
	  for (ie = 0; ie < geom->entities; ie++)
	    {
		pGeo = *(geom->geometries + ie);
		switch (pGeo->type)
		  {
		  case VSHP_POINT:
		      point = (vshp_point_ptr) (pGeo->geometry);
		      *ptr = VSHP_MARK_ENTITY;	/* ENTITY signature */
		      vshp_export_32 (ptr + 1, VSHP_POINT, 1);	/* class POINT */
		      vshp_export_64 (ptr + 5, point->x, 1);	/* X */
		      vshp_export_64 (ptr + 13, point->y, 1);	/* Y */
		      ptr += 21;
		      break;
		  case VSHP_LINESTRING:
		      line = (vshp_linestring_ptr) (pGeo->geometry);
		      *ptr = VSHP_MARK_ENTITY;	/* ENTITY signature */
		      vshp_export_32 (ptr + 1, VSHP_LINESTRING, 1);	/* class LINESTRING */
		      vshp_export_32 (ptr + 5, line->points, 1);	/* # points */
		      ptr += 9;
		      for (iv = 0; iv < line->points; iv++)
			{
			    vshp_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X */
			    vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y */
			    ptr += 16;
			}
		      break;
		  case VSHP_POLYGON:
		      polyg = (vshp_polygon_ptr) (pGeo->geometry);
		      *ptr = VSHP_MARK_ENTITY;	/* ENTITY signature */
		      vshp_export_32 (ptr + 1, VSHP_POLYGON, 1);	/* class POLYGON */
		      vshp_export_32 (ptr + 5, polyg->borders + 1, 1);	/* # rings */
		      line = polyg->external;
		      vshp_export_32 (ptr + 9, line->points, 1);	/* # points - exterior ring */
		      ptr += 13;
		      for (iv = 0; iv < line->points; iv++)
			{
			    vshp_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - exterior ring */
			    vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - exterior ring */
			    ptr += 16;
			}
		      for (ib = 0; ib < polyg->borders; ib++)
			{
			    line = *(polyg->internals + ib);
			    vshp_export_32 (ptr, line->points, 1);	/* # points - interior ring */
			    ptr += 4;
			    for (iv = 0; iv < line->points; iv++)
			      {
				  vshp_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - interior ring */
				  vshp_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - interior ring */
				  ptr += 16;
			      }
			}
		      *ptr = VSHP_MARK_END;	/* END signature */
		      break;
		  };
	    }
	  *ptr = VSHP_MARK_END;	/* END signature */
      }
}

static void
vshp_mbr_geometry (vshp_geometry_ptr geom)
{
/* computes the MBR for this geometry */
    int ie;
    vshp_generic_geometry_ptr pGeo;
    vshp_point_ptr point = NULL;
    vshp_linestring_ptr line = NULL;
    vshp_polygon_ptr polyg = NULL;
    geom->minx = DBL_MAX;
    geom->miny = DBL_MAX;
    geom->maxx = DBL_MIN;
    geom->maxy = DBL_MIN;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case VSHP_POINT:
		point = (vshp_point_ptr) (pGeo->geometry);
		if (point->x < geom->minx)
		    geom->minx = point->x;
		if (point->y < geom->miny)
		    geom->miny = point->y;
		if (point->x > geom->maxx)
		    geom->maxx = point->x;
		if (point->y > geom->maxy)
		    geom->maxy = point->y;
		break;
	    case VSHP_LINESTRING:
		line = (vshp_linestring_ptr) (pGeo->geometry);
		vshp_mbr_linestring (line);
		if (line->minx < geom->minx)
		    geom->minx = line->minx;
		if (line->miny < geom->miny)
		    geom->miny = line->miny;
		if (line->minx > geom->maxx)
		    geom->maxx = line->maxx;
		if (line->miny > geom->maxy)
		    geom->maxy = line->maxy;
		break;
	    case VSHP_POLYGON:
		polyg = (vshp_polygon_ptr) (pGeo->geometry);
		vshp_mbr_polygon (polyg);
		if (polyg->minx < geom->minx)
		    geom->minx = polyg->minx;
		if (polyg->miny < geom->miny)
		    geom->miny = polyg->miny;
		if (polyg->minx > geom->maxx)
		    geom->maxx = polyg->maxx;
		if (polyg->miny > geom->maxy)
		    geom->maxy = polyg->maxy;
		break;
	    };
      }
}

static void
vshp_mbr_linestring (vshp_linestring_ptr line)
{
/* computes the MBR for this linestring */
    int iv;
    double x;
    double y;
    line->minx = DBL_MAX;
    line->miny = DBL_MAX;
    line->maxx = DBL_MIN;
    line->maxy = DBL_MIN;
    for (iv = 0; iv < line->points; iv++)
      {
	  x = (line->geometry + iv)->x;
	  y = (line->geometry + iv)->y;
	  if (x < line->minx)
	      line->minx = x;
	  if (y < line->miny)
	      line->miny = y;
	  if (x > line->maxx)
	      line->maxx = x;
	  if (y > line->maxy)
	      line->maxy = y;
      }
}

static void
vshp_mbr_polygon (vshp_polygon_ptr polyg)
{
/* computes the MBR for this polygon */
    vshp_linestring_ptr line;
    polyg->minx = DBL_MAX;
    polyg->miny = DBL_MAX;
    polyg->maxx = DBL_MIN;
    polyg->maxy = DBL_MIN;
    line = polyg->external;
    vshp_mbr_linestring (line);
    if (line->minx < polyg->minx)
	polyg->minx = line->minx;
    if (line->miny < polyg->miny)
	polyg->miny = line->miny;
    if (line->minx > polyg->maxx)
	polyg->maxx = line->maxx;
    if (line->miny > polyg->maxy)
	polyg->maxy = line->maxy;
}
